from random import randint
from asyncio import Future, get_event_loop, iscoroutine, Event

from pyee import AsyncIOEventEmitter


class DataStoreUnavailable(Exception):
    pass


class User(dict):
    """ Represents a participating user in the chat. """

    def __init__(self, id, robot, **options):
        super().__init__()
        self['id'] = id
        self.update(**options)
        self.__dict__["_robot"] = robot
        if not self.name:
            self.name = str(self.id)

    def __setattr__(self, attr, val):
        self[attr] = val

    def __getattr__(self, attr):
        return self.__dict__.get(attr) or super().get(attr, None)

    def __repr__(self):
        robot = self.__dict__['_robot']
        remains = {k: v for k, v in self.items() if k != 'id'}
        return f"User({self.id}, {robot}, **{remains})"

    def set(self, key, value):
        self._check_datastore_available()
        return self._get_datastore()._set(self._construct(key), value, "users")

    def get(self, key):
        self._check_datastore_available()
        return self._get_datastore()._get(self._construct(key), 'users')

    def _construct(self, key):
        return f"{self.id}+{key}"

    def _get_datastore(self):
        if self._robot:
            return self._robot.datastore

    def _check_datastore_available(self):
        if not self._get_datastore():
            raise DataStoreUnavailable("datastore is not initialized")


class Brain(AsyncIOEventEmitter):
    """ Represents somewhat persistent storage for the robot. """

    def __init__(self, robot, auto_save=True, loop=None):
        super().__init__(loop or get_event_loop())
        #: Enable or disable the automatic saving
        self.auto_save = auto_save

        self._robot = robot
        self.timer = None
        self.data = dict(users=dict(), _private=dict())
        robot.on("running", lambda: self.reset_save_interval(5))

    def set(self, key, value):
        """ Store key-value pair under the private namespace and extend existing
        @data before emitting the 'loaded' event.

        Returns the instance for chaining.
        """

        if isinstance(key, dict):
            pair = key
        else:
            pair = {key: value}
        self.data['_private'].update(pair)

        self.emit('loaded', self.data)
        return self

    def get(self, key):
        """ Get value by key from private namespace in @data or return None if
        not found.
        """
        return self.data['_private'].get(key)

    def remove(self, key):
        """ Remove value by key from the private namespace in @data if it exists
        """

        self.data['_private'].pop(key, None)
        return self

    def save(self):
        """ Emits the 'save' event so that 'brain' scripts can handle persisting
        """
        self.emit("save", self.data)

    def close(self):
        """ Emits the 'close' event so that 'brain' scripts can handle closing.
        """
        if self.timer:
            self.timer.cancel()
        self.save()
        self.emit("close")

    def reset_save_interval(self, seconds):
        """ Reset the interval between save function calls.

        :param seconds: An Integer of seconds between saves.
        """

        def tock():
            if self.auto_save:
                self.save()

        if self.timer:
            self.timer.cancel()
        self.timer = self._loop.call_later(seconds, tock)

    def merge_data(self, data):
        """ Merge keys loaded from a DB against the in memory representation.

        .. caution :: Deeply nested structures don't merge well.
        """

        def renew_ifneed(user):
            if not isinstance(user, User):
                id_ = getattr(user, 'id', "undefined:%s" % id(user))
                return User(id_, self._robot, **getattr(user, 'options', dict()))
            return user

        if isinstance(data, dict):
            self.data.update(data)
            users = self.data['users']
            self.data['users'] = {k:renew_ifneed(u) for k, u in users.items()}

        self.emit("loaded", self.data)

    def users(self):
        """ Get an list of User objects stored in the brain """
        return self.data["users"].values()

    def user_for_id(self, id, **options):
        """ Get a User object given a unique identifier """
        users = self.data['users']
        user = users.get(id, User(id, self._robot, **options))
        users[id] = user

        room = options.get('room')
        if room and (not user.room or user.room != room):
            user = User(id, self._robot, **options)
            users[id] = user

        return user

    def user_for_name(self, name):
        """ Get a User object given a name """
        for user in self.users():
            if str(user.name).lower() == name.lower():
                return user

    def users_for_raw_fuzzy_name(self, fuzzy_name):
        """ Get all users whose names startswith the fuzzy_name. """
        return [u for u in self.users()
                if str(u.name).lower().startwith(fuzzy_name.lower())]

    def users_for_fuzzy_name(self, fuzzy_name):
        """ If fuzzy_name is an exact match for a user, returns a list with just
        that user.  Otherwise, returns a list of all users for which fuzzy_name
        is a raw fuzzy matched in `users_for_raw_fuzzy_name`.
        """
        users = self.users_for_raw_fuzzy_name(fuzzy_name)
        matches = [u for u in users if u.name.lower() == fuzzy_name.lower()]
        return matches if matches else users


class Response:
    ''' Responses are sent to matching listeners.  Messages know about the
    content and user that made the original message, and how to reply back to
    them.

    :param robot: A Robot instance.
    :param message: A Message instance.
    :param match: A Match object from the successful Regex match.
    '''

    def __init__(self, robot, message, match=None):
        self.robot = robot
        self.message = message
        self.match = match
        self.envelope = {
            "room": message.room,
            "user": message.user,
            "message": message
        }

    def send(self, *strings):
        """ Posts a message back to the chat source

        :param strings: One or more strings to be posted.  The order of these
            strings should be kept intact.
        """
        return self._run_middleware('send', *strings, plaintext=True)

    def emote(self, *strings):
        """ Posts an emote back to the chat source

        :param strings: One or more strings to be posted.  The order of these
            strings should be kept intact.
        """
        return self._run_middleware('emote', *strings, plaintext=True)

    def reply(self, *strings):
        """ Posts a message mentioning the current user.

        :param strings: One or more strings to be posted.  The order of these
            strings should be kept intact.
        """
        return self._run_middleware('reply', *strings, plaintext=True)

    def topic(self, *strings):
        """ Posts a topic changing message

        :param strings: One or more strings to set as the topic of the room the
            bot is in.
        """
        return self._run_middleware('topic', *strings, plaintext=True)

    def play(self, *strings):
        """ Posts a sound in the chat source

        :param strings: One or more strings to be posted as sounds to play.
            The order of these strings should be kept intact.
        """
        return self._run_middleware('play', *strings)

    def locked(self, *string):
        """ Posts a message in an unlogged room

        :param strings: One or more strings to be posted.  The order of these
            strings should be kept intact.
        """
        return self._run_middleware('locked', *strings, plaintext=True)

    async def _run_middleware(self, method_name, *strings, plaintext=False):
        ctx = dict(response=self, method=method_name,
                   plaintext=plaintext, strings=list(strings)[:])
        ctx = await self.robot.response_middleware.execute(ctx)
        # XXX: `locked` is not the basic adapter method, only for campfire.
        handle = getattr(self.robot.adapter, method_name)
        coro = handle(self.envelope, *ctx.get('strings', list()))
        if iscoroutine(coro):
            return await coro
        return coro

    def random(self, items):
        """ Picks a random item from the given items.

        :param items: An Sequence of items.
        """
        return items[randint(0, len(items) -1)]

    def finish(self):
        """ Tell the message to stop dispatching to listeners. """
        self.message.finish()


class Middleware:
    def __init__(self, robot):
        self.robot = robot
        self.stack = list()

    def register(self, middleware):
        ''' A generic pipeline component function that can either continue the
        pipeline or interrupt it.

        :param middleware: A callable (async) function that will receive two
            arguments: (context: dict, finish: callable).
        '''

        # we will not check signature and just make sure it is callable.
        if not callable(middleware):
            raise ValueError("middleware should be a callable with 2 arguments.")
        self.stack.append(middleware)

    async def execute(self, context):
        ''' Execute all middleware in order, if the `finish` is not called.

        :async:
        :param context: A dict represents as context that is passed through the
            middleware stack.  When handling errors, this is assumed to have a
            `response` key inside.
        '''
        class Finished(Exception):
            pass

        def finish(*args):
            raise Finished(*args)

        for func in self.stack:
            try:
                coro = func(context, finish)
                if iscoroutine(coro):
                    await coro
            except Finished:
                break
            except Exception as e:
                self.robot.emit("robot", e, context.response)
                break
        return context


class Listener:
    ''' Listeners receive every message from the chat source and decide if they
    want to act on it.
    An identifier should be provided in the options parameter to uniquely
    identify the listener (`id` in options).

    :param robot: A Robot instance
    :param matcher: A `callable` that determines if this listener should trigger
        the handler.
    :param handler: A callable (async) function that is triggered if the incoming
        message matches.
    :param options: Additional parameters keyed on extension name. (optional)
    '''

    regex = None

    def __init__(self, robot, matcher, handler, **options):
        self.robot = robot
        self.matcher = matcher
        self.options = options
        self.handler = handler
        self.options['id'] = self.options.get('id', None)
        if not (callable(matcher) and callable(handler)):
            raise ValueError("Callback function should be callable.")

    async def call(self, message, middleware=None):
        """ Determines if the listener likes the content of the messages.
        If so, a Response built from the given Message is passed through all
        registered middleware and potentially the Listener handler.  Note that
        middleware can intercept the message and prevent the handler from ever
        being executed.

        :async:
        :param message: A Message instance
        :param middleware: Optional Middleware object to execute before the
            listener handled.
        """
        middleware = middleware or Middleware(self.robot)
        match = self.matcher(message)

        if match:
            if self.regex:
                if hasattr(self.regex, 'pattern'):
                    pattern = self.regex.pattern
                else:
                    pattern = self.regex
                    self.robot.logger.info("Unknown expected `re` object.")
                msg = (f"Message '{message}' matched regex `{pattern}`;"
                       f" listener.options = {self.options}")
                self.robot.logger.debug(msg)

            response = Response(self.robot, message, match)
            context = dict(listener=self, response=response)
            await middleware.execute(context)

            try:
                self.robot.logger.debug("Executing listener handler for"
                                        f" Message {message}")
                coro = self.handler(context['response'])
                if iscoroutine(coro):
                    await coro
            except Exception as e:
                self.robot.emit("error", e, context['response'])
            return True

        return False


class TextListener(Listener):
    ''' TextListeners receive every message from the chat source and decide if
    they want to act on it.

    :param robot: A Robot instance
    :param regex: A regular expression that determines if this listener should
        trigger the handler.
    :param handler: A callable (async) function that is triggered if the incoming
        message matches.
    :param options: Additional parameters keyed on extension name. (optional)
    '''

    def __init__(self, robot, regex, handler, **options):
        super().__init__(robot, self._matcher, handler, **options)
        self.regex = re.compile(regex)

    def _matcher(self, message):
        if isinstance(message, TextMessage):
            return message.match(self.regex)



from .plugins import Adapter
from .plugins import TextMessage, CatchAllMessage
from .plugins import EnterMessage, LeaveMessage, TopicMessage
from .robot import Robot

__version__ = '0.2.0'
__all__ = ["Robot", "Adapter", "TextMessage", "EnterMessage", "LeaveMessage",
           "TopicMessage", "CatchAllMessage"]



import re
from asyncio import get_event_loop, Future, iscoroutine, Event

from pyee import AsyncIOEventEmitter

from .core import Response, DataStoreUnavailable


class Adapter(AsyncIOEventEmitter):
    """ An adapter is a specific interface to a chat source for robots. """

    def __init__(self, robot):
        super().__init__(robot._loop)
        self.robot = robot
        self.receive = self.robot.receive  # dispatch

    def send(self, envelope, *strings):
        """ Raw method for sending data back to the chat source.

        :param envelope: An `dict` with message, room and user details.
        :param strings: One or more Strings for each message to send.
        """

    def emote(self, envelope, *strings):
        """ Raw method for sending emote data back to the chat source.
        Defaults as an alias for send.

        :param envelope: An `dict` with message, room and user details.
        :param strings: One or more Strings for each message to send.
        """
        return self.send(envelope, *strings)

    def reply(self, envelope, *strings):
        """ Raw method for building a reply and sending it back to the chat.

        :param envelope: An `dict` with message, room and user details.
        :param strings: One or more Strings for each reply to send.
        """

    def topic(self, envelope, *strings):
        """ Raw method for setting a topic on the chat source.

        :param envelope: An `dict` with message, room and user details.
        :param strings: One or more Strings to set as the topic.
        """

    def play(self, envelope, *strings):
        """ Raw method for playing a sound in the chat source.

        :param envelope: An `dict` with message, room and user details.
        :param strings: One or more Strings for each play message to send.
        """

    def run(self):
        """ Raw method for invoking the bot to run.  Could be a `coroutine`. """

    def close(self):
        """ Raw method for shutting the bot down. """


class Message:
    ''' Represents an incoming message from the chat.

    :param user: A User instance that sent the message.
    '''

    def __init__(self, user, done=False):
        self.user = user
        self.done = done
        self.room = user.room

    def finish(self):
        """ Indicates that no other Listener should be called on this object. """
        self.done = True


class TextMessage(Message):
    ''' Represents an incoming message from the chat.

    :param user: A User instance that sent the message.
    :param text: A message that should be `str`.
    :param id: String of the message ID.
    '''

    def __init__(self, user, text, id):
        super().__init__(user)
        self.text = text
        self.id = id

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.user}, {self.text!r}, {self.id!r})"

    def __str__(self):
        return str(self.text)

    def match(self, regex):
        """ Determines if the message matches the given regex. 

        :param regex: A regular expression to check.
        """
        return regex.match(self.text)


class EnterMessage(Message):
    ''' Represents an incoming user entrance notification.

    :param user: A User instance for the user who entered.
    :param text: always None, meaningless.
    :param id: String of the message ID.
    '''


class LeaveMessage(Message):
    ''' Represents an incoming user exit notification.

    :param user: A User instance for the user who left.
    :param text: always None, meaningless.
    :param id: String of the message ID.
    '''


class TopicMessage(TextMessage):
    ''' Represents an incoming topic change notification. 

    :param user: A User instance for the user who changed the topic.
    :param text: String of the new topic.
    :param id: String of the message ID.
    '''


class CatchAllMessage(Message):
    ''' Represents a message that no matchers matched.

    :param message: The original message.
    '''

    def __init__(self, message):
        super().__init__(message.user)
        self.message = message


class DataStore:
    """ Represents a persistent, database-backed storage for the robot. """

    def __init__(self, robot):
        self._robot = robot

    def set(self, key, value):
        """ Set value for key in the database.  Overwrites existing values
        if present.  Value can be any JSON-serializable type.

        :rtype: asyncio.Future
        """

        return self._set(key, value, 'global')

    async def set_object(self, key, object_key, value):
        """ Assuming `key` represents an dict in the databse, sets its
        `object_key` to `value`.  If `key` isn't already present, it's
        instantiated as an empty dict.

        :async:
        """

        target = await self.get(key) or dict()
        target[object_key] = value
        return await self.set(key, target)

    async def set_array(self, key, value):
        """ Adds the supplied value(s) to the end of the existing array in the
        database marked by `key`.  If `key` isn't already present, it's
        instantiated as an empty list.

        :async:
        """
        target = await self.get(key) or []
        items = target + (value if isinstance(value, list) else [value])
        return await self.set(key, items)

    def get(self, key):
        """ Get value by key if in the database or return `None` if not found.

        :rtype: asyncio.Future
        """
        return self._get(key, 'global')

    async def get_object(self, key, object_key):
        """ Digs inside the object at `key` for a key named `object_key`.
        If `key` isn't already present, or if it doesn't contain an
        `object_key`, returns `None`.

        :async:
        """

        target = await self.get(key) or dict()
        return target.get(object_key)

    def _set(self, key, value, table):
        f = Future()
        f.set_exception(DataStoreUnavailable("Setter called on the abstract class."))
        return f

    def _get(self, key, table):
        f = Future()
        f.set_exception(DataStoreUnavailable("Getter called on the abstract class."))
        return f



import re
import sys
from os import environ
from logging import getLogger, StreamHandler, Formatter
from pathlib import Path
from asyncio import get_event_loop, iscoroutine
from importlib import import_module, util as import_util

from pyee import AsyncIOEventEmitter

from .core import Response, Middleware, Brain, Listener, TextListener
from .plugins import EnterMessage, LeaveMessage, TopicMessage, CatchAllMessage

DEFAULT_ADAPTERS = ["shell"]


def get_logger(level="INFO"):
    fmt = Formatter("[%(asctime)s] [%(module)s.%(funcName)s] [%(levelname)s]::"
                    " %(message)s",
                    datefmt="%Y-%m-%d %H:%M:%S %z")
    ch = StreamHandler()
    ch.setLevel("DEBUG")
    ch.setFormatter(fmt)

    log = getLogger("aiohubot")
    log.addHandler(ch)
    try:
        log.setLevel(level.upper())
    except ValueError:
        log.setLevel("INFO")
        log.error(f"Fail to set log level with errors", exc_info=True)

    return log


class Robot:
    ''' Robots receive messages from a chat source, and dispatch them to
    matching listeners.

    :param adapter: A string that will be passed to `load_adapter` to use.
    :param httpd: Whether to enable the HTTP daemon, disabled by default.
    :param name: A string of the robot name, defaults to Hubot.
    :param alias: A string of the alias of the robot name.
    '''

    def __init__(self, adapter, httpd=False, name="Hubot", alias="", loop=None):
        self.datastore = self.adapter = None
        self.commands, self.listeners, self.error_handlers = [], [], []
        self.listener_middleware = Middleware(self)
        self.response_middleware = Middleware(self)
        self.receive_middleware = Middleware(self)
        self.logger = get_logger(environ.get("HUBOT_LOG_LEVEL", "info"))

        self.name = name
        self.alias = alias
        self._loop = loop or get_event_loop()

        # used in httprouter
        self.server = self.ping_interval_id = None
        self.router = dict()
        if httpd:
            self.setup_httprouter()  # was setupExpress()
        else:
            self.setup_nullrouter()

        self.adapter_name = self.load_adapter(adapter)
        #: delegate to adapter.send
        self.send = self.adapter.send
        #: delegate to adapter.reply
        self.reply = self.adapter.reply

        self.events = AsyncIOEventEmitter()
        #: delegate to events.on
        self.on = self.events.on
        #: delegate to events.emit
        self.emit = self.events.emit
        self.brain = Brain(self)
        self.on("error", self.invoke_error_handlers)

    def __repr__(self):
        name = self.__class__.__name__
        return (f"<{self.__module__}.{name}"
                f" adapter={self.adapter_name!r}"
                f" name={self.name!r} alias={self.alias!r}"
                f" at 0x{id(self):08X}>")

    def listen(self, matcher, handler, **options):
        """ Adds a custom Listener with the provided matcher, handler and other
        options.

        :param matcher: A function returns a truthy value to determines whether
            to call the handler.
        :param handler: A (async) function is called with a Response object if
            the matcher function returns true.
        :param options: additional parameters keyed on extension name.
        """
        self.listeners.append(Listener(self, matcher, handler, **options))

    def hear(self, regex, handler, **options):
        """ Adds a Listener that attempts to match incoming messages based on a
        regular expression.

        :param regex: A Regex that determines if the handler should be called.
        :param handler: A (async) function is called with a Response object.
        :param options: additional parameters keyed on extension name.
        """
        listener = TextListener(self, re.comiple(regex), handler, **options)
        self.listeners.append(listener)

    def response(self, regex, handler, *, flags=0, **options):
        """ Adds a Listener that attempts to match incoming messages directed
        at the robot based on a Regex.  All regexes treat patterns like they
        begin with a '^'

        :param regex: A Regex that determines if the handler should be called.
        :param handler: A (async) function is called with a Response object.
        :param flags: The constants passed to `re.compile`.
        :param options: additional parameters keyed on extension name.
        """
        self.hear(self.response_pattern(regex, flags), handler, **options)

    def response_pattern(self, pattern, flags=0):
        """ Build a regular expression that matches messages addressed directly
        to the robot.

        :param pattern: A string of regex for the message part that follows the
            robot's name/alias.
        :param flags: The constants passed to `re.compile`.
        """
        escape = re.compile(r"[-[\]{}()*+?.,\\^$|#\s]")
        name = escape.sub("\\$&", self.name)
        if pattern.startswith("^"):
            self.logger.warning("Anchors don't work well with response, "
                                "perhaps you want to use `hear`?")
            self.logger.warning(f"the regex is {pattern}")

        if not self.alias:
            return re.compile(fr"^\\s*[@]?{name}[:,]?\\s*(?:{pattern})", flags)

        alias = escape("\\$&", self.alias)
        # XXX: it seems not need to return in different order in python
        if len(name) > len(alias):
            x, y = name, alias
            _pattern = fr"^\\s*[@]?(?:{x}[:,]?|{y}[:,]?)\\s*(?:{pattern})"
        else:
            _pattern = fr"^\\s*[@]?(?:{y}[:,]?|{x}[:,]?)\\s*(?:{pattern})"
        return re.compile(_pattern, flags)

    def enter(self, handler, **options):
        """ Adds a Listener that triggers when anyone enters the room.

        :param handler: A (async) function is called with a Response object.
        :param options: additional parameters keyed on extension name.
        """
        self.listen(lambda m: isinstance(m, EnterMessage), handler, **options)

    def leave(self, handler, **options):
        """ Adds a Listener that triggers when anyone leaves the room.

        :param handler: A (async) function is called with a Response object.
        :param options: additional parameters keyed on extension name.
        """
        self.listen(lambda m: isinstance(m, LeaveMessage), handler, **options)

    def topic(self, handler, **options):
        """ Adds a Listener that triggers when anyone changes the topic.

        :param handler: A (async) function is called with a Response object.
        :param options: additional parameters keyed on extension name.
        """
        self.listen(lambda m: isinstance(m, TopicMessage), handler, **options)

    def error(self, handler):
        """ Adds an error handler when an uncaught exception or user emitted
        error event occurs.

        :param handler: A (async) function is called with the error object.
        """
        self.error_handlers.append(handler)

    async def invoke_error_handlers(self, err, res=None):
        """ Calls and passes any registered error handlers for unhandled
        exceptions or user emitted error events.

        :async:
        :param err: An Exception object.
        :param res: An optional Response object that generated the error.
        """
        self.logger.exception(err)
        for hdlr in self.error_handlers:
            try:
                coro = hdlr(err, res)
                if iscoroutine(coro):
                    await coro
            except Exception:
                msg = "exception when invoking error handler:"
                self.logger.error(msg, exc_info=True)

    def catch_all(self, handler, **options):
        """ Adds a Listener that triggers when no other text matchers match.

        :param handler: A (async) function is called with a Response object.
        :param options: additional parameters keyed on extension name.
        """
        def _check(m):
            return isinstance(m, CatchAllMessage)

        def hdlr(msg):
            msg.message = msg.message.message
            return handler(msg)

        self.listen(_check, hdlr, **options)

    def middleware_listener(self, middleware):
        """ Registers new middleware for execution after matching but before
        listener handled.

        :param middleware: A callable (async) function that will receive two
            arguments: (context: dict, finish: callable).
        """
        self.listener_middleware.register(middleware)

    def middleware_response(self, middleware):
        """ Registers new middleware for execution as a response to any message
        is being sent.

        :param middleware: A callable (async) function that will receive two
            arguments: (context: dict, finish: callable).
        """
        self.response_middleware.register(middleware)

    def middleware_receive(self, middleware):
        """ Registers new middleware for execution before matching.

        :param middleware: A callable (async) function that will receive two
            arguments: (context: dict, finish: callable).
        """
        self.receive_middleware.register(middleware)

    async def receive(self, message):
        """ Passes the given message to any interested Listeners after running
        receive middleware.

        :param message: A Message instance.  Listeners can flag this message as
            'done' to prevent further execution.
        """
        context = dict(response=Response(self, message))
        await self.receive_middleware.execute(context)
        executed = False
        for listener in self.listeners:
            try:
                executed = await listener.call(context['response'].message,
                                               self.listener_middleware)
                if message.done:
                    break
            except Exception as e:
                self.emit("error", e, Response(self, context['response'].message))
        else:
            msg = context['response'].message
            if not isinstance(msg, CatchAllMessage) and not executed:
                self.logger.debug("No listeners executed; failling back to catch-all")
                executed = await self.receive(CatchAllMessage(msg))
        return executed

    def load_file(self, filepath):
        """ Loads a file in path.

        :param filepath: A string of full filepath in the filesystem.
        """
        fpath = Path(filepath).absolute()
        try:
            spec = import_util.spec_from_file_location(fpath.stem, fpath)
            module = import_util.module_from_spec(spec)
            spec.loader.exec_module(module)
            if hasattr(module, "use"):
                if callable(module.use):
                    module.use(self)
                    self.parse_help(module.__doc__)
                else:
                    self.logger.warning(f"Expected `use` in {fpath} isn't callable.")
            else:
                self.logger.warning(f"{fpath} not expose expected `use` function.")
        except Exception:
            self.logger.error(f"Unable to load {fpath}:", exc_info=True)
            sys.exit(1)

    def load(self, path):
        """ Loads every python script in the given path.

        :param path: A string path on the filesystem.
        """
        self.logger.debug(f"Loading scripts from {path}")
        p = Path(path)
        if p.is_dir():
            for f in p.iterdir():
                if f.is_file() and f.name.endswith(".py"):
                    self.load_file(f)
        elif p.is_file() and p.name.endswith(".py"):
            self.load_file(p)

    def load_external_scripts(self, packages):
        """ Load scripts from packages specified in the `external-scripts.json`
        file.

        :param packages: An list of packages containing hubot scripts to load.
            Or A dict that key is the module name, and value is the additional
            parameter to be passed into `use` function as second argument.
        """
        self.logger.debug("Loading external-scripts")
        try:
            if isinstance(packages, list):
                return [import_module(p).use(self) for p in packages]
            elif isinstance(packages, dict):
                return [import_module(p).use(self, v) for p, v in packages.items()]
        except Exception:
            self.logger.error(f"Error loading scripts", exc_info=True)
            sys.exit(1)

    def setup_httprouter(self):
        raise NotImplementedError("Not support yet.")

    def setup_nullrouter(self):
        """ Setup an empty router object. """
        def _warn():
            return self.logger.warning(msg)
        msg = ("A script has tried registering a HTTP route"
               " while the HTTP server is disabled with --disabled-httpd.")
        self.router = dict(get=_warn, post=_warn, put=_warn, delete=_warn)

    def load_adapter(self, adapter):
        """ Load the adapter Hubot is going to use.

        :param adapter: A String of the adapter name to use.  It supports the
            'pkg.adapters.module:initial_function' syntax.  The section after
            `:` is on demend.  If not set, `module.use` will be called.
        :return: The last module name if the `initial_function` is not set,
            otherwise return the initial_function name.
        """
        self.logger.debug(f"loading adapter {adapter}")
        try:
            if adapter in DEFAULT_ADAPTERS:
                mod_name = "aiohubot.adapters.%s" % adapter
                has_attrs = attrs = ""
            else:
                mod_name, has_attrs, attrs = adapter.partition(":")

            module = import_module(mod_name)
            if has_attrs:
                init_adap = module
                parts = attrs.split(".")
                for part in parts:
                    init_adap = getattr(init_adap, part)
                name = init_adap.__name__
            else:
                init_adap = getattr(module, 'use')
                name = mod_name.rsplit(".", 1)[-1]
            self.adapter = init_adap(self)
        except Exception as e:
            self.logger.error(f"Cannot load adapter {adapter} - {e}")
            self.logger.exception(e)
            sys.exit(1)

        return name

    def help_commands(self):
        """ Help commands for Running Scripts. """
        return sorted(self.commands)

    def parse_help(self, document):
        pass

    def message_room(self, room, *strings):
        """ A helper send function to message a room that the robot is in.

        :param room: String designating the room to message.
        :param strings: One or more Strings for each message to send.
        """
        envelope = dict(room=room)
        self.adapter.send(envelope, *strings)

    def run(self):
        """ Kick off the event loop for the adapter. """
        self.emit("running")
        coro = self.adapter.run()
        if iscoroutine(coro):
            self._loop.create_task(coro)
        try:
            self._loop.run_forever()
        except KeyboardInterrupt:
            self.shutdown()

    def shutdown(self):
        """ Gracefully shutdown the robot process. """

        if self.ping_interval_id:
            self.ping_interval_id.cancel()
        self.adapter.close()
        if self.server:
            self.server.close()
        self.brain.close()
